#ifdef FSBENCH

#include "ext2.h"
#include <linux/string.h>
#include <linux/file.h>
#include <linux/sched.h>
#include <linux/buffer_head.h>

void init_fsbench_stat(void);
void print_fsbench_stat(void);
void update_factor(int num, unsigned int factor);

unsigned int fsb_factors[FSB_NUM];
unsigned int fsb_num_calls[FSB_NUM];
unsigned long long fsb_op_time[FSB_NUM];
unsigned long long fsb_wait_time[FSB_NUM];

void update_factor(int num, unsigned int factor) {
	fsb_factors[num] = factor;
}

#ifdef __ia64__
#define RDTSC(qp) \
	do { \
		unsigned long result; \
		__asm__ __volatile__("mov %0=ar.itc" : "=r" (result) :: "memory"); \
		qp = result; \
	} while (0);

#else

#define RDTSC(qp) \
	do { \
		unsigned long lowPart, highPart; \
		__asm__ __volatile__("rdtsc" : "=a" (lowPart), "=d" (highPart)); \
		qp = (((unsigned long long) highPart) << 32) | lowPart; \
	} while (0);

#endif

void init_fsbench_stat(void) {
	int i;

	for (i = 0 ; i < FSB_NUM ; i++) {
		fsb_num_calls[i] = 0;
		fsb_op_time[i] = 0;
		fsb_wait_time[i] = 0;
	}

	printk("fsbench initialized.\n");
}

struct file fastcall *fget(unsigned int fd)
{
	struct file *file;
	struct files_struct *files = current->files;

	spin_lock(&files->file_lock);
	file = fcheck_files(files, fd);
	if (file)
		get_file(file);
	spin_unlock(&files->file_lock);
	return file;
}

void print_fsbench_stat(void) {
	int i;
	char buf[80];
	struct file *wfilp = fget(2);
	mm_segment_t oldfs;

	for(i = 0 ; i < FSB_NUM ; i++) {
		memset(buf, 0, sizeof(buf));
		snprintf(buf, sizeof(buf), "%d(%d): num_calls: %d , op_time: %llu , wait_time: %llu\n", i, fsb_factors[i], fsb_num_calls[i], fsb_op_time[i], fsb_wait_time[i]);
		oldfs = get_fs();
		set_fs(KERNEL_DS);
		wfilp->f_op->write(wfilp, buf, strlen(buf), &wfilp->f_pos);
		set_fs(oldfs);
	}
	fput(wfilp);
}

ssize_t fsbench_do_sync_read(struct file *filp, char __user *buf, size_t count, loff_t *ppos) {
	ssize_t ret;
	unsigned long long bef, aft, goal, now;

	RDTSC(bef);
	ret = do_sync_read(filp, buf, count, ppos);
	RDTSC(aft);
	if (!fsb_factors[FSB_READ])
		return ret;

	now = aft;
	goal = now + ((aft - bef) << fsb_factors[FSB_READ]) - (aft - bef);

	while (now <= goal) {
		if (fsb_factors[FSB_SCH])
			schedule();
		RDTSC(now);
	}

	//number of ops, total time spent, total time waited
	fsb_num_calls[FSB_READ]++;
	fsb_op_time[FSB_READ] += (aft - bef);
	fsb_wait_time[FSB_READ] += (now - aft);

	return ret;
}

int fsbench_writepage(struct page *page, struct writeback_control *wbc) {
	int ret;
	unsigned long long bef, aft, goal, now;

	RDTSC(bef);
	ret = ext2_writepage(page, wbc);
	RDTSC(aft);
	if (!fsb_factors[FSB_WRITEPAGE])
		return ret;

	now = aft;
	goal = now + ((aft - bef) << fsb_factors[FSB_WRITEPAGE]) - (aft - bef);

	while (now <= goal) {
		if (fsb_factors[FSB_SCH])
			schedule();
		RDTSC(now);
	}

	//number of ops, total time spent, total time waited
	fsb_num_calls[FSB_WRITEPAGE]++;
	fsb_op_time[FSB_WRITEPAGE] += (aft - bef);
	fsb_wait_time[FSB_WRITEPAGE] += (now - aft);

	return ret;
}

int fsbench_writepages(struct address_space *mapping, struct writeback_control *wbc) {
	int ret;
	unsigned long long bef, aft, goal, now;

	RDTSC(bef);
	ret = ext2_writepages(mapping, wbc);
	RDTSC(aft);
	if (!fsb_factors[FSB_WRITEPAGES])
		return ret;

	now = aft;
	goal = now + ((aft - bef) << fsb_factors[FSB_WRITEPAGES]) - (aft - bef);

	while (now <= goal) {
		if (fsb_factors[FSB_SCH])
			schedule();
		RDTSC(now);
	}

	//number of ops, total time spent, total time waited
	fsb_num_calls[FSB_WRITEPAGES]++;
	fsb_op_time[FSB_WRITEPAGES] += (aft - bef);
	fsb_wait_time[FSB_WRITEPAGES] += (now - aft);

	return ret;
}

int fsbench_prepare_write(struct file *file, struct page *page, unsigned from, unsigned to) {
	int ret;
	unsigned long long bef, aft, goal, now;

	RDTSC(bef);
	ret = ext2_prepare_write(file, page, from, to);
	RDTSC(aft);
	if (!fsb_factors[FSB_PREP_WRITE])
		return ret;

	now = aft;
	goal = now + ((aft - bef) << fsb_factors[FSB_PREP_WRITE]) - (aft - bef);

	while (now <= goal) {
		if (fsb_factors[FSB_SCH])
			schedule();
		RDTSC(now);
	}

	//number of ops, total time spent, total time waited
	fsb_num_calls[FSB_PREP_WRITE]++;
	fsb_op_time[FSB_PREP_WRITE] += (aft - bef);
	fsb_wait_time[FSB_PREP_WRITE] += (now - aft);

	return ret;
}

int fsbench_commit_write(struct file *file, struct page *page, unsigned from, unsigned to) {
	int ret;
	unsigned long long bef, aft, goal, now;

	RDTSC(bef);
	ret = generic_commit_write(file, page, from, to);
	RDTSC(aft);
	if (!fsb_factors[FSB_COMM_WRITE])
		return ret;

	now = aft;
	goal = now + ((aft - bef) << fsb_factors[FSB_COMM_WRITE]) - (aft - bef);

	while (now <= goal) {
		if (fsb_factors[FSB_SCH])
			schedule();
		RDTSC(now);
	}

	//number of ops, total time spent, total time waited
	fsb_num_calls[FSB_COMM_WRITE]++;
	fsb_op_time[FSB_COMM_WRITE] += (aft - bef);
	fsb_wait_time[FSB_COMM_WRITE] += (now - aft);

	return ret;
}

struct dentry *fsbench_lookup(struct inode * dir, struct dentry *dentry, struct nameidata *nd) {
	struct dentry *ret;
	unsigned long long bef, aft, goal, now;

	RDTSC(bef);
	ret = ext2_lookup(dir, dentry, nd);
	RDTSC(aft);
	if (!fsb_factors[FSB_LOOKUP])
		return ret;

	now = aft;
	goal = now + ((aft - bef) << fsb_factors[FSB_LOOKUP]) - (aft - bef);

	while (now <= goal) {
		if (fsb_factors[FSB_SCH])
			schedule();
		RDTSC(now);
	}

	//number of ops, total time spent, total time waited
	fsb_num_calls[FSB_LOOKUP]++;
	fsb_op_time[FSB_LOOKUP] += (aft - bef);
	fsb_wait_time[FSB_LOOKUP] += (now - aft);

	return ret;
}

#endif //FSBENCH
